﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using Arction.Wpf.Charting;
using Arction.Wpf.Charting.Axes;
using Arction.Wpf.Charting.SeriesXY;
using Arction.Wpf.Charting.Views.ViewXY;

namespace CsLightningChartBug
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        private const int SegmentCount = 4;
        private const int SeriesCount = 7;
        private const int SampleCount = 512;
        private static readonly TimeSpan SamplingDuration = TimeSpan.FromMilliseconds(20);

        private LightningChartUltimate lcu;
        private List<SampleDataSeries> allSeries;
        private List<AxisY> allAxes;
        private Task dataFeedTask;
        private volatile bool isFeedingData;

        public MainWindow()
        {
            InitializeComponent();

            this.isFeedingData = true;
            this.lcu = new LightningChartUltimate();

            PrepareChart();
            StartFeedingData();

            //!!!!! Bug inducing deferred view
            AttachToViewDeferred();
            //If we comment deferred attach and uncomment direct attach then everything works
            //AttachToView();
        }

        private void PrepareChart()
        {
            lcu.BeginUpdate();
            var view = lcu.ViewXY;

            view.YAxes.Clear();
            view.AxisLayout.Segments.Clear();

            allSeries = new List<SampleDataSeries>();
            allAxes = new List<AxisY>();

            view.AxisLayout.YAxesLayout = YAxesLayout.Segmented;

            for (int i = 0; i < SegmentCount; ++i)
            {
                view.AxisLayout.Segments.Add(new YAxisSegment(view));
            }

            for (int i = 0; i < SeriesCount; i++)
            {
                var color = GetSeriesColor(i);

                var axis = new AxisY(view);
                axis.SegmentIndex = i;
                axis.AxisColor = color;
                axis.Minimum = -0.1;
                axis.Maximum = 1.1;

                var series = new SampleDataSeries(view, view.XAxes[0], axis);
                series.LineStyle.Color = color;
                series.SamplingFrequency = 1.0 / (SamplingDuration.TotalSeconds / SampleCount);

                if (i < SegmentCount)
                {
                    view.YAxes.Add(axis);
                    view.SampleDataSeries.Add(series);
                }

                allSeries.Add(series);
                allAxes.Add(axis);
            }

            view.XAxes[0].ScrollMode = XAxisScrollMode.Scrolling;
            view.XAxes[0].Maximum = 0.4;
            
            lcu.EndUpdate();
        }

        private Color GetSeriesColor(int i)
        {
            switch (i % 8)
            {
                case 0:
                    return Colors.Red;
                case 1:
                    return Colors.Green;
                case 2:
                    return Colors.Blue;
                case 3:
                    return Colors.Orange;
                case 4:
                    return Colors.Cyan;
                case 5:
                    return Colors.Magenta;
                case 6:
                    return Colors.Goldenrod;
                case 7:
                    return Colors.Chartreuse;
                default:
                    return Colors.Black;
            }
        }


        private void StartFeedingData()
        {
            this.dataFeedTask = Task.Run(() => FeedDataLoop());
        }

        private void FeedDataLoop()
        {
            var random = new Random();

            double[][] samples = new double[SeriesCount][];

            for (int i = 0; i < SeriesCount; i++)
            {
                samples[i] = new double[SampleCount];

                for (int j = 0; j < SampleCount; j++)
                {
                    samples[i][j] = random.NextDouble();
                }
            }


            while (isFeedingData)
            {
                int iteration = 0;

                try
                {
                    FeedData(samples, iteration);
                }
                catch
                { }

                Thread.Sleep(1);
                ++iteration;
            }
        }

        private void FeedData(double[][] samples, int iteration)
        {
            if (Dispatcher.CheckAccess())
            {
                FeedDataCore(samples, iteration);
            }
            else
            {
                Dispatcher.Invoke(() => FeedDataCore(samples, iteration));
            }
        }

        private void FeedDataCore(double[][] samples, int iteration)
        {
            lcu.BeginUpdate();

            for (int i = 0; i < SeriesCount; i++)
            {
                double[] seriesSamples = samples[(iteration + i) % SeriesCount];

                allSeries[i].AddSamples(seriesSamples, false);
            }

            lcu.ViewXY.XAxes[0].ScrollPosition = lcu.ViewXY.XAxes[0].ScrollPosition + SamplingDuration.TotalSeconds;

            lcu.EndUpdate();
        }


        protected override void OnClosing(System.ComponentModel.CancelEventArgs e)
        {
            if (dataFeedTask != null)
            {
                e.Cancel = true;
                this.isFeedingData = false;
                this.dataFeedTask.ContinueWith(t =>
                {
                    dataFeedTask = null;
                    Dispatcher.Invoke(() => Close());
                });
            }
            else 
            {
                base.OnClosing(e);
            }
        }

        private void AttachToViewDeferred()
        {
            Task.Delay(3000)
                .ContinueWith(t => {
                    Dispatcher.Invoke(() => AttachToView());
                });
        }

        private void AttachToView()
        {
            Content = lcu;
            MakeAllSeriesVisible();
        }

        private void MakeAllSeriesVisible()
        {
            lcu.BeginUpdate();

            var view = lcu.ViewXY;

            for (int i = SegmentCount; i < SeriesCount; i++)
            {
                view.AxisLayout.Segments.Add(new YAxisSegment(view));
                view.YAxes.Add(allAxes[i]);
                view.SampleDataSeries.Add(allSeries[i]);
            }

            lcu.EndUpdate();
        }
    }
}
